open Fof_parse;;
open Cnf;;
open Fof;;
open Arglean;;

let c1 = ref 0;;
let c2 = ref 0;;
let c3 = ref 0;;
let infer = ref 0;;
let depth = ref 0;;

let predord (p1, t1) (p2, t2) =
  match compare (abs p1) (abs p2) with
    0 -> (match compare p1 p2 with 0 -> compare t1 t2
         | x -> x)
  | x -> x;;

let rec lexord ((_, t1) as l1) ((_, t2) as l2) =
  match predord l1 l2 with
    0 -> lexord_tl (t1, t2)
  | x -> x
and lexord_tl = function
    [], [] -> 0
  | (h1 :: t1, h2 :: t2) -> (match lexord_t (h1, h2) with
      0 -> lexord_tl (t1, t2)
    | x -> x)
  | _ -> failwith "lexord_tl"
and lexord_t = function
    (A(f1, l1), A(f2, l2)) -> (match compare f1 f2 with
      0 -> lexord_tl (l1, l2)
    | x -> x)
  | _ -> -2;;
let lexords x y = match lexord x y with -2 -> 0 | x -> x;;

let rec sizet acc = function V _ -> 1 + acc | A (_, l) -> List.fold_left sizet (acc + 1) l;;
let sizel acc (_, l) = List.fold_left sizet acc l;;
let sizec l = List.fold_left sizel 0 l;;

let db = Hashtbl.create 1000;;
let db2 = ref [];;

let queue = ref Im.empty;;
let enqueue pri v =
  try let (ovpush, ovpop) = Im.find pri !queue in queue := Im.add pri (v :: ovpush, ovpop) !queue
  with Not_found -> queue := Im.add pri ([], [v]) !queue;;
let rec dequeue () =
  try match Im.min_binding !queue with
    (pri, ([], [])) -> queue := Im.remove pri !queue; dequeue ()
  | (pri, (l, [])) -> (match List.rev l with h :: t -> queue := Im.add pri ([], t) !queue; h | _ -> failwith "dequeue")
  | (pri, (ovpush, h :: t)) -> queue := Im.add pri (ovpush, t) !queue; h
  with Not_found -> failwith "no unprocessed clauses";;

let weight = function [] -> 0 | h :: t -> sizel (List.length t) h;;
let enqueue (c1, c2, w) = enqueue (sizec c1 + sizec c2) (c1, c2, w);;

let rec rev_uniq acc = function
  x::(y::_ as t) -> rev_uniq (if Pervasives.compare x y = 0 then acc else x :: acc) t
| [x] -> (x :: acc)
| [] -> acc;;

let uniq l = List.rev (rev_uniq [] l);;

let rec merge cmp l1 l2 =
  match l1, l2 with
  | [], l2 -> l2
  | l1, [] -> l1
  | h1 :: t1, h2 :: t2 ->
      if cmp h1 h2 <= 0
      then h1 :: merge cmp t1 l2
      else h2 :: merge cmp l1 t2;;

let resolves np tl rest maxv1 =
  let db = try Hashtbl.find db np with Not_found -> Dtree.empty_dt in
  let totry = Dtree.unifs_nojl [] db tl in
  let rec resolves (tl2, rest2, maxv2) =
    try
      let r1, r2 = unify_rename2 maxv1 tl tl2 rest rest2 in
      if r1 = [] && r2 = [] then raise Exit else
      enqueue (r1, r2, maxv1 + maxv2);
      incr infer
    with Unify -> ()
  in
  List.iter resolves totry;;

let factor p tl maxvar = function
  [] -> false
| (((p2, tl2) :: rest2) as rest1) ->
    if p2 <> p then false else try
      let s = unify_list [] tl tl2 in
      let nrest1 = if s = empty_sub then rest1 else List.map (fun (p, tl) -> p, List.map inst_tm tl) rest1 in
      enqueue (nrest1, [], maxvar); true
    with Unify -> false;;

let rec process1 ((p, tl) as l1) rest maxvar =
  if rest <> [] && List.hd rest = l1 then process1 (List.hd rest) (List.tl rest) maxvar else
  let db' = try Hashtbl.find db p with Not_found -> Dtree.empty_dt in
  if Dtree.mem (tl, rest, maxvar) db' tl then () else begin
    (* print_clause ((p, tl) :: rest); Format.printf "\n";*)
    ignore (factor p tl maxvar rest);
    resolves (-p) tl rest maxvar;
    Hashtbl.replace db p ((*Dtree.update_jl*) (Dtree.insert (tl, rest, maxvar) db' tl));
    db2 := (l1 :: rest, maxvar) :: !db2
  end

let rec process () =
  let (c1, c2, _) = dequeue () in
  (* rename_unbound_clause *)
  match rename_unbound_clause (uniq (merge compare c1 c2)) with ([], _) -> raise Exit
  | (l1 :: rest, maxvar) -> process1 l1 rest maxvar; process ();;

let ores file =
  let mat = file_mat false !def true file in
  let mat = List.map (List.sort compare) mat in
  (*List.iter (fun x -> Format.print_char '%'; print_clause x; Format.print_char '\n') mat; Format.printf "%%--\n%!";*)
  let mat = List.map rename_unbound_clause mat in
  let ins = function
    | [], _ -> raise Exit
    | c, maxvar -> enqueue (c, [], maxvar) in
  List.iter ins mat;
  process ();;

(*  let max_var cl = 1 + List.fold_left (fun sf (_, t) -> List.fold_left max_var sf t) (-1) cl in*)


exception Alarm;;
exception CleanExit;;
let alarm _ = raise Alarm;;
let cleanexit _ = raise CleanExit;;
Sys.signal Sys.sigint (Sys.Signal_handle cleanexit);;
Sys.signal Sys.sigterm (Sys.Signal_handle cleanexit);;
Sys.signal Sys.sigalrm (Sys.Signal_handle alarm);;
let maxsec i f a = ignore (Unix.alarm i); try f a; ignore (Unix.alarm 0); () with Alarm -> ();;
at_exit (fun () ->
Format.printf "%% Inf: %i Depth: %i c1: %i c2: %i c3: %i\n" !infer !depth !c1 !c2 !c3);;

if !tosolve <> "" then try ores !tosolve
  with Exit -> Format.printf "\n%% SZS status Theorem\n%!"
     | Failure x -> Format.printf "\n%% SZS status Error\n%%%s\n%!" x
     | CleanExit -> Format.printf "\n%% SZS status Unknown\n%!"
     | Parsing.Parse_error -> Format.printf "\n%% SZS status Error\n%%Parse_error\n%!";;
